import { CommonModule } from '@angular/common';
import { EventEmitter, Injectable, ApplicationRef, ChangeDetectorRef, SimpleChange, Directive, ViewContainerRef, Inject, Injector, SkipSelf, Input, NgModule } from '@angular/core';
import { isObservable, Subject, BehaviorSubject, Observable, ReplaySubject } from 'rxjs';
import { takeUntil } from 'rxjs/operators';

/**
 * @fileoverview added by tsickle
 * Generated from: lib/ngx-bind-io.config.ts
 * @suppress {checkTypes,constantProperty,extraRequire,missingOverride,missingReturn,unusedPrivateMembers,uselessCode} checked by tsc
 */
/** @type {?} */
const defaultNgxBindIOConfig = {
    debug: false
};
/** @type {?} */
const NGX_BIND_IO_CONFIG = 'NgxBindIOConfig';

/**
 * @fileoverview added by tsickle
 * Generated from: lib/utils/bind-io-metadata-utils.ts
 * @suppress {checkTypes,constantProperty,extraRequire,missingOverride,missingReturn,unusedPrivateMembers,uselessCode} checked by tsc
 */
/** @type {?} */
const __BIND_IO__ = '__bindIO__';
/** @type {?} */
const __ORIGINAL_NGONCHANGES__ = '__originalNgOnChanges__';
/** @type {?} */
const __ORIGINAL_NGONDESTROY__ = '__originalNgOnDestroy__';
/**
 * @param {?} component
 * @return {?}
 */
function getBindIOMetadata(component) {
    /** @type {?} */
    const emptyMetadata = {
        asHost: {
            subjects: {}
        },
        asInner: {
            subscriptions: {},
            manualInputs: undefined,
            manualOutputs: undefined
        }
    };
    if (component) {
        if (!component[__BIND_IO__]) {
            ((/** @type {?} */ (component[__BIND_IO__]))) = emptyMetadata;
        }
        return (/** @type {?} */ (component[__BIND_IO__]));
    }
    return emptyMetadata;
}

/**
 * @fileoverview added by tsickle
 * Generated from: lib/utils/components-utils.ts
 * @suppress {checkTypes,constantProperty,extraRequire,missingOverride,missingReturn,unusedPrivateMembers,uselessCode} checked by tsc
 */
/**
 * @param {?} component
 * @param {?} rule
 * @param {?=} maxLevel
 * @return {?}
 */
function collectKeys(component, rule, maxLevel) {
    if (maxLevel !== undefined) {
        maxLevel--;
        if (maxLevel === 0) {
            return [];
        }
    }
    /** @type {?} */
    let keys = component ? Object.keys(component).filter((/**
     * @param {?} propName
     * @return {?}
     */
    propName => rule(component, propName))) : [];
    if (component && component.__proto__) {
        keys = [...keys, ...collectKeys(component.__proto__, rule, maxLevel)];
    }
    return keys;
}
/**
 * @param {?} directive
 * @param {?} existsKeys
 * @return {?}
 */
function removeKeysUsedInAttributes(directive, existsKeys) {
    /** @type {?} */
    const viewContainerRef = directive.viewContainerRef;
    if (viewContainerRef &&
        viewContainerRef.element &&
        viewContainerRef.element.nativeElement &&
        viewContainerRef.element.nativeElement.attributes) {
        /** @type {?} */
        const attributes = (/** @type {?} */ (viewContainerRef.element.nativeElement.attributes));
        /** @type {?} */
        const filtered = existsKeys.filter((/**
         * @param {?} existKey
         * @return {?}
         */
        existKey => {
            /** @type {?} */
            let founded = false;
            for (let i = 0; i < attributes.length; i++) {
                /** @type {?} */
                const name = attributes
                    .item(i)
                    .name.replace(new RegExp('-', 'g'), '')
                    .toUpperCase();
                if (existKey.replace(new RegExp('-', 'g'), '').toUpperCase() === name) {
                    founded = true;
                }
            }
            return !founded;
        }));
        return filtered;
    }
    return existsKeys;
}
/**
 * @param {?} directive
 * @param {?} existsKeys
 * @return {?}
 */
function removeKeysManualBindedOutputs(directive, existsKeys) {
    if (directive.ignoreKeysManualBinded) {
        return existsKeys;
    }
    /** @type {?} */
    const filtered = existsKeys.filter((/**
     * @param {?} innerKey
     * @return {?}
     */
    innerKey => {
        /** @type {?} */
        const innerFiltered = (getBindIOMetadata(directive.innerComponent).asInner.manualOutputs
            ? Object.keys(getBindIOMetadata(directive.innerComponent).asInner.manualOutputs)
            : []).filter((/**
         * @param {?} outputName
         * @return {?}
         */
        outputName => outputName.toUpperCase() === innerKey.toUpperCase())).length === 0;
        return innerFiltered;
    }));
    return filtered;
}
/**
 * @param {?} directive
 * @param {?} existsKeys
 * @return {?}
 */
function removeKeysManualBindedInputs(directive, existsKeys) {
    if (directive.ignoreKeysManualBinded) {
        return existsKeys;
    }
    /** @type {?} */
    const filtered = existsKeys.filter((/**
     * @param {?} innerKey
     * @return {?}
     */
    innerKey => {
        /** @type {?} */
        const innerFiltered = (getBindIOMetadata(directive.innerComponent).asInner.manualInputs
            ? Object.keys(getBindIOMetadata(directive.innerComponent).asInner.manualInputs)
            : []).filter((/**
         * @param {?} inputName
         * @return {?}
         */
        inputName => inputName.toUpperCase() === innerKey.toUpperCase())).length === 0;
        return innerFiltered;
    }));
    return filtered;
}
/**
 * @param {?} directive
 * @param {?} existsKeys
 * @return {?}
 */
function removeKeysNotAllowedConstants(directive, existsKeys) {
    /** @type {?} */
    const constants = [
        __BIND_IO__,
        __ORIGINAL_NGONCHANGES__,
        __ORIGINAL_NGONDESTROY__,
        'constructor',
        'ngOnChanges',
        'ngOnDestroy'
    ];
    /** @type {?} */
    const filtered = existsKeys.filter((/**
     * @param {?} key
     * @return {?}
     */
    key => constants.indexOf(key) === -1));
    return filtered;
}

/**
 * @fileoverview added by tsickle
 * Generated from: lib/utils/property-utils.ts
 * @suppress {checkTypes,constantProperty,extraRequire,missingOverride,missingReturn,unusedPrivateMembers,uselessCode} checked by tsc
 */
/** @type {?} */
const values = new WeakMap();
/**
 * @param {?} instance
 * @return {?}
 */
function valueMap(instance) {
    /** @type {?} */
    let _valueMap = values.get(instance);
    if (!_valueMap) {
        _valueMap = new Map();
        values.set(instance, _valueMap);
    }
    return _valueMap;
}
/**
 * @param {?} target
 * @param {?} propertyKey
 * @param {?=} callback
 * @return {?}
 */
function redefineSimpleProperty(target, propertyKey, callback) {
    Object.defineProperty(target, propertyKey, {
        configurable: true,
        enumerable: true,
        /**
         * @param {?} value
         * @return {?}
         */
        set(value) {
            valueMap(this).set(propertyKey, value);
            if (callback) {
                callback(value);
            }
        },
        /**
         * @return {?}
         */
        get() {
            return valueMap(this).get(propertyKey);
        }
    });
}
/**
 * @param {?} target
 * @param {?} propertyKey
 * @param {?} descriptor
 * @param {?=} callback
 * @return {?}
 */
function redefineAccessorProperty(target, propertyKey, descriptor, callback) {
    Object.defineProperty(target, propertyKey, {
        configurable: descriptor.configurable,
        enumerable: descriptor.enumerable,
        /**
         * @param {?} value
         * @return {?}
         */
        set(value) {
            if (!descriptor.set) {
                throw new Error(`Property ${propertyKey} doesn't have a setter and cannot be written`);
            }
            descriptor.set.call(this, value);
            if (!descriptor.get) {
                callback(value);
            }
            else {
                callback(descriptor.get());
            }
        },
        /**
         * @return {?}
         */
        get() {
            if (!descriptor.get) {
                throw new Error(`Property ${propertyKey} doesn't have a getter and cannot be read`);
            }
            return descriptor.get.call(this);
        }
    });
}
/**
 * @param {?} target
 * @param {?} propertyKey
 * @return {?}
 */
function getPropDescriptor(target, propertyKey) {
    /** @type {?} */
    let originalDescriptor = Reflect.getOwnPropertyDescriptor(target, propertyKey);
    /** @type {?} */
    let setter = originalDescriptor && originalDescriptor.set;
    /** @type {?} */
    let getter = originalDescriptor && originalDescriptor.get;
    /** @type {?} */
    const value = originalDescriptor && Object.getOwnPropertyDescriptor(originalDescriptor, 'value')
        ? originalDescriptor.value
        : target[propertyKey];
    if (!setter || ((/** @type {?} */ (target))).__lookupSetter__(propertyKey)) {
        setter = ((/** @type {?} */ (target))).__lookupSetter__(propertyKey);
    }
    if (!getter || ((/** @type {?} */ (target))).__lookupGetter__(propertyKey)) {
        getter = ((/** @type {?} */ (target))).__lookupGetter__(propertyKey);
    }
    if (!originalDescriptor && setter && getter) {
        originalDescriptor = {
            configurable: true,
            enumerable: true,
            set: setter.bind(target),
            get: getter.bind(target)
        };
    }
    return {
        originalDescriptor: originalDescriptor,
        value,
        setter,
        getter
    };
}

/**
 * @fileoverview added by tsickle
 * Generated from: lib/utils/utils.ts
 * @suppress {checkTypes,constantProperty,extraRequire,missingOverride,missingReturn,unusedPrivateMembers,uselessCode} checked by tsc
 */
/**
 * @param {?} functionToCheck
 * @return {?}
 */
function isFunction(functionToCheck) {
    /** @type {?} */
    const result = functionToCheck && {}.toString.call(functionToCheck) === '[object Function]';
    return result;
}

/**
 * @fileoverview added by tsickle
 * Generated from: lib/services/ngx-bind-inputs.service.ts
 * @suppress {checkTypes,constantProperty,extraRequire,missingOverride,missingReturn,unusedPrivateMembers,uselessCode} checked by tsc
 */
class NgxBindInputsService {
    /**
     * BindInputs Directive
     * @param {?} directive
     * @return {?}
     */
    bindInputs(directive) {
        /** @type {?} */
        const inputs = directive.inputs;
        const { includeInputs, excludeInputs } = this.getIncludesAndExcludes(directive);
        inputs.hostKeys
            .filter((/**
         * @param {?} hostKey
         * @return {?}
         */
        hostKey => (includeInputs.length === 0 && excludeInputs.indexOf(hostKey.toUpperCase()) === -1) ||
            (includeInputs.length !== 0 && includeInputs.indexOf(hostKey.toUpperCase()) !== -1)))
            .forEach((/**
         * @param {?} hostKey
         * @return {?}
         */
        hostKey => {
            inputs.innerKeys
                .filter((/**
             * @param {?} innerKey
             * @return {?}
             */
            innerKey => (includeInputs.length === 0 && excludeInputs.indexOf(innerKey.toUpperCase()) === -1) ||
                (includeInputs.length !== 0 && includeInputs.indexOf(innerKey.toUpperCase()) !== -1)))
                .forEach((/**
             * @param {?} innerKey
             * @return {?}
             */
            innerKey => {
                if (this.checkInputToBind(directive, hostKey, innerKey)) {
                    this.bindInput(directive, hostKey, innerKey);
                }
            }));
        }));
    }
    /**
     * @param {?} directive
     * @return {?}
     */
    bindObservableInputs(directive) {
        /** @type {?} */
        const inputs = directive.inputs;
        const { includeInputs, excludeInputs } = this.getIncludesAndExcludes(directive);
        inputs.hostKeys
            .filter((/**
         * @param {?} hostKey
         * @return {?}
         */
        hostKey => (includeInputs.length === 0 && excludeInputs.indexOf(hostKey.toUpperCase()) === -1) ||
            (includeInputs.length !== 0 && includeInputs.indexOf(hostKey.toUpperCase()) !== -1)))
            .forEach((/**
         * @param {?} hostKey
         * @return {?}
         */
        hostKey => {
            inputs.innerKeys
                .filter((/**
             * @param {?} innerKey
             * @return {?}
             */
            innerKey => (includeInputs.length === 0 && excludeInputs.indexOf(innerKey.toUpperCase()) === -1) ||
                (includeInputs.length !== 0 && includeInputs.indexOf(innerKey.toUpperCase()) !== -1)))
                .forEach((/**
             * @param {?} innerKey
             * @return {?}
             */
            innerKey => {
                if (this.checkObservableInputToBind(directive, hostKey, innerKey)) {
                    this.bindObservableInput(directive, hostKey, innerKey);
                }
            }));
        }));
    }
    /**
     * Inputs
     * @param {?} directive
     * @param {?} hostKey
     * @param {?} innerKey
     * @return {?}
     */
    checkKeyNameToInputBind(directive, hostKey, innerKey) {
        return hostKey === innerKey && hostKey[0] !== '_';
    }
    /**
     * @param {?} directive
     * @param {?} hostKey
     * @param {?} innerKey
     * @return {?}
     */
    checkInputToBind(directive, hostKey, innerKey) {
        /** @type {?} */
        const hostValue = getPropDescriptor(directive.hostComponent, hostKey).value;
        return (directive.usedInputs[hostKey] === undefined &&
            !isObservable(hostValue) &&
            this.checkKeyNameToInputBind(directive, hostKey, innerKey));
    }
    /**
     * @param {?} directive
     * @param {?} hostKey
     * @param {?} innerKey
     * @return {?}
     */
    bindInput(directive, hostKey, innerKey) {
        directive.usedInputs[hostKey] = innerKey;
        /** @type {?} */
        const descriptor = getPropDescriptor(directive.hostComponent, hostKey);
        /** @type {?} */
        const currentValue = descriptor.value;
        /** @type {?} */
        const currentHostSubject = getBindIOMetadata(directive.hostComponent).asHost.subjects[hostKey];
        if (!currentHostSubject) {
            getBindIOMetadata(directive.hostComponent).asHost.subjects[hostKey] = new Subject();
            delete directive.hostComponent[hostKey];
            if (descriptor !== undefined && descriptor.setter !== undefined && descriptor.getter !== undefined) {
                redefineAccessorProperty(directive.hostComponent, hostKey, descriptor.originalDescriptor, (/**
                 * @param {?} newValue
                 * @return {?}
                 */
                (newValue) => getBindIOMetadata(directive.hostComponent).asHost.subjects[hostKey].next(newValue)));
            }
            else {
                redefineSimpleProperty(directive.hostComponent, hostKey, (/**
                 * @param {?} newValue
                 * @return {?}
                 */
                (newValue) => getBindIOMetadata(directive.hostComponent).asHost.subjects[hostKey].next(newValue)));
            }
        }
        getBindIOMetadata(directive.hostComponent)
            .asHost.subjects[hostKey].pipe(takeUntil(directive.destroyed$))
            .subscribe((/**
         * @param {?} value
         * @return {?}
         */
        value => directive.bindValue(innerKey, value)));
        try {
            if (!currentHostSubject) {
                directive.hostComponent[hostKey] = currentValue;
            }
            else {
                if (!directive.ignoreKeysManualBinded) {
                    directive.bindValue(innerKey, currentValue);
                }
            }
        }
        catch (error) { }
    }
    /**
     * Observable Inputs
     * @param {?} directive
     * @param {?} hostKey
     * @param {?} innerKey
     * @return {?}
     */
    checkKeyNameToObservableInputBind(directive, hostKey, innerKey) {
        return hostKey === `${innerKey}$` && hostKey[0] !== '_';
    }
    /**
     * @param {?} directive
     * @param {?} hostKey
     * @param {?} innerKey
     * @return {?}
     */
    checkObservableInputToBind(directive, hostKey, innerKey) {
        /** @type {?} */
        const descriptor = getPropDescriptor(directive.hostComponent, hostKey);
        /** @type {?} */
        const hostValue = descriptor.value;
        return (directive.usedInputs[hostKey] === undefined &&
            isObservable(hostValue) &&
            this.checkKeyNameToObservableInputBind(directive, hostKey, innerKey));
    }
    /**
     * @param {?} directive
     * @param {?} hostKey
     * @param {?} innerKey
     * @return {?}
     */
    bindObservableInput(directive, hostKey, innerKey) {
        directive.usedInputs[hostKey] = innerKey;
        /** @type {?} */
        const descriptor = getPropDescriptor(directive.hostComponent, hostKey);
        /** @type {?} */
        const isBehaviorSubject = descriptor.value instanceof BehaviorSubject;
        /** @type {?} */
        const currentValue = descriptor.value;
        /** @type {?} */
        let behaviorSubjectValue;
        if (descriptor.value instanceof BehaviorSubject) {
            behaviorSubjectValue = descriptor.value.getValue();
        }
        /** @type {?} */
        const currentHostSubject = getBindIOMetadata(directive.hostComponent).asHost.subjects[hostKey];
        if (!currentHostSubject) {
            getBindIOMetadata(directive.hostComponent).asHost.subjects[hostKey] = new Subject();
            delete directive.hostComponent[hostKey];
            if (descriptor !== undefined && descriptor.setter !== undefined && descriptor.getter !== undefined) {
                redefineAccessorProperty(directive.hostComponent, hostKey, descriptor.originalDescriptor, (/**
                 * @param {?} newValue
                 * @return {?}
                 */
                (newValue) => getBindIOMetadata(directive.hostComponent).asHost.subjects[hostKey].next(newValue)));
            }
            else {
                redefineSimpleProperty(directive.hostComponent, hostKey, (/**
                 * @param {?} newValue
                 * @return {?}
                 */
                (newValue) => getBindIOMetadata(directive.hostComponent).asHost.subjects[hostKey].next(newValue)));
            }
        }
        getBindIOMetadata(directive.hostComponent)
            .asHost.subjects[hostKey].pipe(takeUntil(directive.destroyed$))
            .subscribe((/**
         * @param {?} newValue
         * @return {?}
         */
        newValue => {
            if (newValue instanceof Observable) {
                if (getBindIOMetadata(directive.innerComponent).asInner.subscriptions[hostKey]) {
                    getBindIOMetadata(directive.innerComponent).asInner.subscriptions[hostKey].unsubscribe();
                }
                getBindIOMetadata(directive.innerComponent).asInner.subscriptions[hostKey] = newValue
                    .pipe(takeUntil(directive.destroyed$))
                    .subscribe((/**
                 * @param {?} value
                 * @return {?}
                 */
                value => directive.bindValue(innerKey, value)));
            }
        }));
        try {
            directive.hostComponent[hostKey] = currentValue;
        }
        catch (error) { }
        if (isBehaviorSubject) {
            currentValue.next(behaviorSubjectValue);
        }
    }
    /**
     * Utils
     * @param {?} directive
     * @return {?}
     */
    getInputs(directive) {
        const { hostPropDescriptorKeys, innerPropDescriptorKeys } = this.getInputsFromPropDescriptor(directive);
        const { hostIvyKeys, innerIvyKeys } = this.getInputsFromIvy(directive);
        /** @type {?} */
        let innerKeys = [...innerPropDescriptorKeys, ...innerIvyKeys];
        /** @type {?} */
        let hostKeys = [...hostPropDescriptorKeys, ...hostIvyKeys];
        hostKeys = [
            ...hostKeys,
            ...hostKeys
                .map((/**
             * @param {?} hostKey
             * @return {?}
             */
            hostKey => hostKey + '$'))
                .filter((/**
             * @param {?} hostKey
             * @return {?}
             */
            hostKey => hostKeys.indexOf(hostKey) === -1 &&
                (directive.hostComponent[hostKey] instanceof ReplaySubject ||
                    directive.hostComponent[hostKey] instanceof Subject ||
                    directive.hostComponent[hostKey] instanceof BehaviorSubject)))
        ];
        innerKeys = removeKeysManualBindedInputs(directive, removeKeysUsedInAttributes(directive, innerKeys));
        hostKeys = removeKeysUsedInAttributes(directive, hostKeys);
        /** @type {?} */
        const inputs = {
            hostKeys: removeKeysNotAllowedConstants(directive, hostKeys),
            innerKeys: removeKeysNotAllowedConstants(directive, innerKeys)
        };
        return inputs;
    }
    /**
     * @private
     * @param {?} directive
     * @return {?}
     */
    getInputsFromPropDescriptor(directive) {
        /** @type {?} */
        const innerPropDescriptorKeys = directive.innerComponent
            ? [
                ...Object.keys(directive.innerComponent).filter((/**
                 * @param {?} innerKey
                 * @return {?}
                 */
                innerKey => !(getPropDescriptor(directive.innerComponent, innerKey).value instanceof EventEmitter))),
                ...collectKeys(directive.innerComponent.__proto__, (/**
                 * @param {?} cmp
                 * @param {?} innerKey
                 * @return {?}
                 */
                (cmp, innerKey) => !(getPropDescriptor(cmp, innerKey).value instanceof EventEmitter)), 10)
            ]
            : [];
        /** @type {?} */
        const hostPropDescriptorKeys = collectKeys(directive.hostComponent, (/**
         * @param {?} cmp
         * @param {?} hostKey
         * @return {?}
         */
        (cmp, hostKey) => !isFunction(getPropDescriptor(directive.hostComponent, hostKey).value)), 10);
        return { hostPropDescriptorKeys, innerPropDescriptorKeys };
    }
    /**
     * @private
     * @param {?} directive
     * @return {?}
     */
    getInputsFromIvy(directive) {
        var _a, _b, _c, _d, _e, _f, _g, _h;
        /** @type {?} */
        const innerIvyKeys = directive.innerComponent
            ? [
                ...Object.keys(((_d = (_c = (_b = (_a = directive.innerComponent) === null || _a === void 0 ? void 0 : _a.__proto__) === null || _b === void 0 ? void 0 : _b.constructor) === null || _c === void 0 ? void 0 : _c.ɵcmp) === null || _d === void 0 ? void 0 : _d.inputs) || {}).filter((/**
                 * @param {?} key
                 * @return {?}
                 */
                key => !(getPropDescriptor(directive.innerComponent, key).value instanceof EventEmitter)))
            ]
            : [];
        /** @type {?} */
        const hostIvyKeys = directive.hostComponent
            ? [
                ...Object.keys(((_h = (_g = (_f = (_e = directive.hostComponent) === null || _e === void 0 ? void 0 : _e.__proto__) === null || _f === void 0 ? void 0 : _f.constructor) === null || _g === void 0 ? void 0 : _g.ɵcmp) === null || _h === void 0 ? void 0 : _h.inputs) || {}).filter((/**
                 * @param {?} key
                 * @return {?}
                 */
                key => !(getPropDescriptor(directive.hostComponent, key).value instanceof EventEmitter)))
            ]
            : [];
        return { hostIvyKeys, innerIvyKeys };
    }
    /**
     * @param {?} directive
     * @return {?}
     */
    getIncludesAndExcludes(directive) {
        /** @type {?} */
        const exclude = Array.isArray(directive.excludeInputs) ? directive.excludeInputs : [directive.excludeInputs];
        /** @type {?} */
        const include = Array.isArray(directive.includeInputs) ? directive.includeInputs : [directive.includeInputs];
        /** @type {?} */
        const includeIO = !directive.includeIO
            ? []
            : Array.isArray(directive.includeIO)
                ? directive.includeIO
                : [directive.includeIO];
        /** @type {?} */
        const excludeIO = !directive.excludeIO
            ? []
            : Array.isArray(directive.excludeIO)
                ? directive.excludeIO
                : [directive.excludeIO];
        /** @type {?} */
        const excludeInputs = [...exclude, ...excludeIO].map((/**
         * @param {?} excludeKey
         * @return {?}
         */
        excludeKey => excludeKey.toUpperCase()));
        /** @type {?} */
        const includeInputs = [...include, ...includeIO].map((/**
         * @param {?} includeKey
         * @return {?}
         */
        includeKey => includeKey.toUpperCase()));
        return { includeInputs, excludeInputs };
    }
}
NgxBindInputsService.decorators = [
    { type: Injectable }
];

/**
 * @fileoverview added by tsickle
 * Generated from: lib/services/ngx-bind-io-debug.service.ts
 * @suppress {checkTypes,constantProperty,extraRequire,missingOverride,missingReturn,unusedPrivateMembers,uselessCode} checked by tsc
 */
class NgxBindIODebugService {
    /**
     * @param {?} _ngxBindInputsService
     */
    constructor(_ngxBindInputsService) {
        this._ngxBindInputsService = _ngxBindInputsService;
    }
    /**
     * @param {?} directive
     * @param {?} debug
     * @return {?}
     */
    showDebugInfo(directive, debug) {
        /** @type {?} */
        let notExistsOutputs = [];
        /** @type {?} */
        let notExistsInputs = [];
        if (directive.outputs &&
            directive.innerComponent &&
            directive.innerComponent.__proto__ &&
            directive.innerComponent.__proto__.constructor &&
            directive.innerComponent.__proto__.constructor.ngBaseDef &&
            directive.innerComponent.__proto__.constructor.ngBaseDef.outputs) {
            /** @type {?} */
            const ngBaseDefOutputs = Object.keys(directive.innerComponent.__proto__.constructor.ngBaseDef.outputs);
            notExistsOutputs = removeKeysManualBindedOutputs(directive, removeKeysUsedInAttributes(directive, ngBaseDefOutputs)).filter((/**
             * @param {?} ngBaseDefOutput
             * @return {?}
             */
            ngBaseDefOutput => directive.outputs.innerKeys.indexOf(ngBaseDefOutput) === -1));
        }
        if (directive.inputs &&
            directive.innerComponent &&
            directive.innerComponent.__proto__ &&
            directive.innerComponent.__proto__.constructor &&
            directive.innerComponent.__proto__.constructor.ngBaseDef &&
            directive.innerComponent.__proto__.constructor.ngBaseDef.inputs) {
            /** @type {?} */
            const ngBaseDefInputs = Object.keys(directive.innerComponent.__proto__.constructor.ngBaseDef.inputs);
            notExistsInputs = removeKeysManualBindedInputs(directive, removeKeysUsedInAttributes(directive, ngBaseDefInputs)).filter((/**
             * @param {?} ngBaseDefInput
             * @return {?}
             */
            ngBaseDefInput => directive.inputs.innerKeys.indexOf(ngBaseDefInput) === -1));
        }
        if (debug || notExistsOutputs.length > 0 || notExistsInputs.length > 0) {
            if (debug) {
                console.group('NgxBindIO: debug');
            }
            else {
                console.group('NgxBindIO: warning');
            }
            console.log('Host component:', directive.hostComponent.__proto__.constructor.name, directive.hostComponent);
            console.log('Inner component:', directive.innerComponent.__proto__.constructor.name, directive.innerComponent);
            console.log('Directive:', directive);
            if (directive.usedOutputs) {
                console.log('Outputs maping:', Object.keys(directive.usedOutputs).map((/**
                 * @param {?} hostKey
                 * @return {?}
                 */
                hostKey => `(${directive.usedOutputs[hostKey]})="${hostKey}($event)"`)));
            }
            if (notExistsOutputs.length > 0) {
                console.log('Not initialized outputs:', notExistsOutputs);
                console.log('Not initialized outputs (text):', JSON.stringify(notExistsOutputs));
            }
            if (directive.usedInputs) {
                console.log('Inputs maping:', Object.keys(directive.usedInputs).map((/**
                 * @param {?} hostKey
                 * @return {?}
                 */
                hostKey => this._ngxBindInputsService.checkKeyNameToObservableInputBind(directive, hostKey, directive.usedInputs[hostKey])
                    ? `[${directive.usedInputs[hostKey]}]="${hostKey} | async"`
                    : `[${directive.usedInputs[hostKey]}]="${hostKey}"`)));
            }
            if (notExistsInputs.length > 0) {
                console.log('Not initialized inputs:', notExistsInputs);
                console.log('Not initialized inputs (text):', JSON.stringify(notExistsInputs));
            }
            console.groupEnd();
        }
    }
}
NgxBindIODebugService.decorators = [
    { type: Injectable }
];
/** @nocollapse */
NgxBindIODebugService.ctorParameters = () => [
    { type: NgxBindInputsService }
];
if (false) {
    /**
     * @type {?}
     * @private
     */
    NgxBindIODebugService.prototype._ngxBindInputsService;
}

/**
 * @fileoverview added by tsickle
 * Generated from: lib/directives/bind-inputs.directive.ts
 * @suppress {checkTypes,constantProperty,extraRequire,missingOverride,missingReturn,unusedPrivateMembers,uselessCode} checked by tsc
 */
class BindInputsDirective {
    /**
     * @param {?} viewContainerRef
     * @param {?} _ngxBindIOConfig
     * @param {?} _ngxBindInputsService
     * @param {?} _ngxBindIODebugService
     * @param {?} _parentInjector
     * @param {?} _detectorRef
     */
    constructor(viewContainerRef, _ngxBindIOConfig, _ngxBindInputsService, _ngxBindIODebugService, _parentInjector, _detectorRef) {
        this.viewContainerRef = viewContainerRef;
        this._ngxBindIOConfig = _ngxBindIOConfig;
        this._ngxBindInputsService = _ngxBindInputsService;
        this._ngxBindIODebugService = _ngxBindIODebugService;
        this._parentInjector = _parentInjector;
        this._detectorRef = _detectorRef;
        this.excludeInputs = [];
        this.includeInputs = [];
        this.usedInputs = {};
        this.destroyed$ = new Subject();
        this.innerSimpleChanges = {};
    }
    /**
     * @param {?} simpleChanges
     * @return {?}
     */
    ngOnChanges(simpleChanges) {
        this.detectComponents();
    }
    /**
     * @return {?}
     */
    ngAfterContentInit() {
        this.detectComponents();
        this.bindAll();
    }
    /**
     * @return {?}
     */
    bindAll() {
        this.inputs = this._ngxBindInputsService.getInputs(this);
        this._ngxBindInputsService.bindInputs(this);
        this._ngxBindInputsService.bindObservableInputs(this);
        this._ngxBindIODebugService.showDebugInfo(this, this.debugIsActive());
    }
    /**
     * @return {?}
     */
    ngOnDestroy() {
        this.destroyed$.next(true);
        this.destroyed$.complete();
    }
    /**
     * @param {?} innerKey
     * @param {?} value
     * @return {?}
     */
    bindValue(innerKey, value) {
        /** @type {?} */
        const previousValue = this.innerComponent[innerKey];
        this.innerComponent[innerKey] = value;
        if (this._detectorRef && typeof this._detectorRef.markForCheck === 'function') {
            this._detectorRef.markForCheck();
        }
        if (this._detectorRef instanceof ApplicationRef) {
            setTimeout((/**
             * @return {?}
             */
            () => ((/** @type {?} */ (((/** @type {?} */ (this._detectorRef)))))).tick()), 50);
        }
        if (!this._detectorRef) {
            Object.keys(this.innerComponent).forEach((/**
             * @param {?} key
             * @return {?}
             */
            key => {
                if (this.innerComponent[key] instanceof ChangeDetectorRef) {
                    this.innerComponent[key].markForCheck();
                }
            }));
        }
        if (typeof this.innerComponent['ngOnChanges'] === 'function') {
            /** @type {?} */
            const simpleChange = new SimpleChange(previousValue, value, this.innerSimpleChanges[innerKey] === undefined);
            this.innerComponent['ngOnChanges']({ [innerKey]: simpleChange });
            this.innerSimpleChanges[innerKey] = simpleChange;
        }
    }
    /**
     * @return {?}
     */
    detectComponents() {
        var _a, _b;
        if (this.viewContainerRef && !this.innerComponent && !this.hostComponent) {
            // ivy
            this.innerComponent = (_a = ((/** @type {?} */ (this._detectorRef)))) === null || _a === void 0 ? void 0 : _a.context;
            this.hostComponent =
                this._parentInjector &&
                    (((_b = ((/** @type {?} */ (this._parentInjector)))) === null || _b === void 0 ? void 0 : _b._lView) || []).filter((/**
                     * @param {?} item
                     * @return {?}
                     */
                    item => item && item.__ngContext__))[1];
            // if not ivy try old logic
            if (!this.innerComponent) {
                this.innerComponent = this.viewContainerRef['_data'].componentView.component;
            }
            // if not ivy try old logic
            if (!this.hostComponent) {
                this.hostComponent = ((/** @type {?} */ (this.viewContainerRef)))._view.context;
                if (this.hostComponent.$implicit !== undefined) {
                    this.hostComponent = ((/** @type {?} */ (this.viewContainerRef)))._view.component;
                }
            }
        }
    }
    /**
     * @return {?}
     */
    debugIsActive() {
        return this._ngxBindIOConfig.debug || (this.bindInputs && this.bindInputs.debug);
    }
}
BindInputsDirective.decorators = [
    { type: Directive, args: [{
                selector: '[bindInputs]'
            },] }
];
/** @nocollapse */
BindInputsDirective.ctorParameters = () => [
    { type: ViewContainerRef },
    { type: undefined, decorators: [{ type: Inject, args: [NGX_BIND_IO_CONFIG,] }] },
    { type: NgxBindInputsService },
    { type: NgxBindIODebugService },
    { type: Injector, decorators: [{ type: SkipSelf }] },
    { type: ChangeDetectorRef }
];
BindInputsDirective.propDecorators = {
    bindInputs: [{ type: Input }],
    excludeInputs: [{ type: Input }],
    includeInputs: [{ type: Input }]
};
if (false) {
    /** @type {?} */
    BindInputsDirective.prototype.bindInputs;
    /** @type {?} */
    BindInputsDirective.prototype.excludeInputs;
    /** @type {?} */
    BindInputsDirective.prototype.includeInputs;
    /** @type {?} */
    BindInputsDirective.prototype.innerComponent;
    /** @type {?} */
    BindInputsDirective.prototype.hostComponent;
    /** @type {?} */
    BindInputsDirective.prototype.inputs;
    /** @type {?} */
    BindInputsDirective.prototype.usedInputs;
    /** @type {?} */
    BindInputsDirective.prototype.destroyed$;
    /** @type {?} */
    BindInputsDirective.prototype.innerSimpleChanges;
    /** @type {?} */
    BindInputsDirective.prototype.viewContainerRef;
    /**
     * @type {?}
     * @private
     */
    BindInputsDirective.prototype._ngxBindIOConfig;
    /**
     * @type {?}
     * @private
     */
    BindInputsDirective.prototype._ngxBindInputsService;
    /**
     * @type {?}
     * @private
     */
    BindInputsDirective.prototype._ngxBindIODebugService;
    /**
     * @type {?}
     * @private
     */
    BindInputsDirective.prototype._parentInjector;
    /**
     * @type {?}
     * @private
     */
    BindInputsDirective.prototype._detectorRef;
}

/**
 * @fileoverview added by tsickle
 * Generated from: lib/services/ngx-bind-outputs.service.ts
 * @suppress {checkTypes,constantProperty,extraRequire,missingOverride,missingReturn,unusedPrivateMembers,uselessCode} checked by tsc
 */
class NgxBindOutputsService {
    /**
     * BindOutputs Directive
     * @param {?} directive
     * @return {?}
     */
    bindOutputs(directive) {
        /** @type {?} */
        const outputs = directive.outputs;
        const { includeOutputs, excludeOutputs } = this.getIncludesAndExcludes(directive);
        outputs.hostKeys
            .filter((/**
         * @param {?} hostKey
         * @return {?}
         */
        hostKey => (includeOutputs.length === 0 && excludeOutputs.indexOf(hostKey.toUpperCase()) === -1) ||
            (includeOutputs.length !== 0 && includeOutputs.indexOf(hostKey.toUpperCase()) !== -1)))
            .forEach((/**
         * @param {?} hostKey
         * @return {?}
         */
        hostKey => {
            outputs.innerKeys
                .filter((/**
             * @param {?} innerKey
             * @return {?}
             */
            innerKey => (includeOutputs.length === 0 && excludeOutputs.indexOf(innerKey.toUpperCase()) === -1) ||
                (includeOutputs.length !== 0 && includeOutputs.indexOf(innerKey.toUpperCase()) !== -1)))
                .forEach((/**
             * @param {?} innerKey
             * @return {?}
             */
            innerKey => {
                if (this.checkOutputToBind(directive, hostKey, innerKey)) {
                    this.bindOutput(directive, hostKey, innerKey);
                }
            }));
        }));
    }
    /**
     * Outputs
     * @param {?} directive
     * @param {?} hostKey
     * @param {?} innerKey
     * @return {?}
     */
    checkKeyNameToOutputBind(directive, hostKey, innerKey) {
        /** @type {?} */
        const outputs = directive.outputs;
        /** @type {?} */
        const keyWithFirstUpperLetter = innerKey.length > 0 ? innerKey.charAt(0).toUpperCase() + innerKey.substr(1) : innerKey;
        return ((hostKey === `on${keyWithFirstUpperLetter}` &&
            outputs.hostKeys.indexOf(`on${keyWithFirstUpperLetter}Click`) === -1) ||
            hostKey === `on${keyWithFirstUpperLetter}Click`);
    }
    /**
     * @param {?} directive
     * @param {?} hostKey
     * @param {?} innerKey
     * @return {?}
     */
    checkOutputToBind(directive, hostKey, innerKey) {
        /** @type {?} */
        const value = getPropDescriptor(directive.innerComponent, innerKey).value;
        return (directive.usedOutputs[hostKey] === undefined &&
            value instanceof EventEmitter &&
            this.checkKeyNameToOutputBind(directive, hostKey, innerKey));
    }
    /**
     * @param {?} directive
     * @param {?} hostKey
     * @param {?} innerKey
     * @return {?}
     */
    bindOutput(directive, hostKey, innerKey) {
        directive.usedOutputs[hostKey] = innerKey;
        directive.innerComponent[innerKey].subscribe((/**
         * @param {?} value
         * @return {?}
         */
        value => directive.hostComponent[hostKey](value)));
    }
    /**
     * Utils
     * @param {?} directive
     * @return {?}
     */
    getOutputs(directive) {
        const { innerPropDescriptorKeys, hostPropDescriptorKeys } = this.getOutputsFromPropDescriptor(directive);
        const { hostIvyKeys, innerIvyKeys } = this.getIvyOutputs(directive);
        /** @type {?} */
        let innerKeys = [...innerPropDescriptorKeys, ...innerIvyKeys];
        /** @type {?} */
        let hostKeys = [...hostPropDescriptorKeys, ...hostIvyKeys];
        innerKeys = removeKeysManualBindedOutputs(directive, removeKeysUsedInAttributes(directive, innerKeys));
        hostKeys = removeKeysUsedInAttributes(directive, hostKeys);
        /** @type {?} */
        const outputs = {
            hostKeys: removeKeysNotAllowedConstants(directive, hostKeys),
            innerKeys: removeKeysNotAllowedConstants(directive, innerKeys)
        };
        if (directive.innerComponent.name === 'BasicBindIOComponent') {
            console.log(directive.innerComponent.name, { directive, outputs });
        }
        return outputs;
    }
    /**
     * @private
     * @param {?} directive
     * @return {?}
     */
    getOutputsFromPropDescriptor(directive) {
        /** @type {?} */
        const hostPropDescriptorKeys = [
            ...Object.keys(directive.hostComponent).filter((/**
             * @param {?} hostKey
             * @return {?}
             */
            hostKey => isFunction(directive.hostComponent[hostKey]))),
            ...collectKeys(directive.hostComponent.__proto__, (/**
             * @param {?} cmp
             * @param {?} hostKey
             * @return {?}
             */
            (cmp, hostKey) => isFunction(getPropDescriptor(cmp, hostKey).value)), 10)
        ];
        /** @type {?} */
        const innerPropDescriptorKeys = directive.innerComponent
            ? [
                ...Object.keys(directive.innerComponent).filter((/**
                 * @param {?} innerKey
                 * @return {?}
                 */
                innerKey => getPropDescriptor(directive.innerComponent, innerKey).value instanceof EventEmitter)),
                ...collectKeys(directive.innerComponent.__proto__, (/**
                 * @param {?} cmp
                 * @param {?} innerKey
                 * @return {?}
                 */
                (cmp, innerKey) => getPropDescriptor(cmp, innerKey).value instanceof EventEmitter), 10)
            ]
            : [];
        return { innerPropDescriptorKeys, hostPropDescriptorKeys };
    }
    /**
     * @private
     * @param {?} directive
     * @return {?}
     */
    getIvyOutputs(directive) {
        var _a, _b, _c, _d, _e, _f, _g, _h;
        /** @type {?} */
        const hostIvyKeys = directive.hostComponent
            ? [
                ...Object.keys(((_d = (_c = (_b = (_a = directive.hostComponent) === null || _a === void 0 ? void 0 : _a.__proto__) === null || _b === void 0 ? void 0 : _b.constructor) === null || _c === void 0 ? void 0 : _c.ɵcmp) === null || _d === void 0 ? void 0 : _d.outputs) || {}).filter((/**
                 * @param {?} key
                 * @return {?}
                 */
                key => getPropDescriptor(directive.hostComponent, key).value instanceof EventEmitter))
            ]
            : [];
        /** @type {?} */
        const innerIvyKeys = [
            ...Object.keys(((_h = (_g = (_f = (_e = directive.innerComponent) === null || _e === void 0 ? void 0 : _e.__proto__) === null || _f === void 0 ? void 0 : _f.constructor) === null || _g === void 0 ? void 0 : _g.ɵcmp) === null || _h === void 0 ? void 0 : _h.outputs) || {}).filter((/**
             * @param {?} key
             * @return {?}
             */
            key => getPropDescriptor(directive.innerComponent, key).value instanceof EventEmitter))
        ];
        return { innerIvyKeys, hostIvyKeys };
    }
    /**
     * @param {?} directive
     * @return {?}
     */
    getIncludesAndExcludes(directive) {
        /** @type {?} */
        const exclude = Array.isArray(directive.excludeOutputs) ? directive.excludeOutputs : [directive.excludeOutputs];
        /** @type {?} */
        const include = Array.isArray(directive.includeOutputs) ? directive.includeOutputs : [directive.includeOutputs];
        /** @type {?} */
        const includeIO = !directive.includeIO
            ? []
            : Array.isArray(directive.includeIO)
                ? directive.includeIO
                : [directive.includeIO];
        /** @type {?} */
        const excludeIO = !directive.excludeIO
            ? []
            : Array.isArray(directive.excludeIO)
                ? directive.excludeIO
                : [directive.excludeIO];
        /** @type {?} */
        const excludeOutputs = [...exclude, ...excludeIO].map((/**
         * @param {?} exludeKey
         * @return {?}
         */
        exludeKey => exludeKey.toUpperCase()));
        /** @type {?} */
        const includeOutputs = [...include, ...includeIO].map((/**
         * @param {?} includeKey
         * @return {?}
         */
        includeKey => includeKey.toUpperCase()));
        return { includeOutputs, excludeOutputs };
    }
}
NgxBindOutputsService.decorators = [
    { type: Injectable }
];

/**
 * @fileoverview added by tsickle
 * Generated from: lib/directives/bind-io.directive.ts
 * @suppress {checkTypes,constantProperty,extraRequire,missingOverride,missingReturn,unusedPrivateMembers,uselessCode} checked by tsc
 */
class BindIODirective {
    /**
     * @param {?} viewContainerRef
     * @param {?} _ngxBindIOConfig
     * @param {?} _ngxBindInputsService
     * @param {?} _ngxBindOutputsService
     * @param {?} _ngxBindIODebugService
     * @param {?} _parentInjector
     * @param {?} _detectorRef
     */
    constructor(viewContainerRef, _ngxBindIOConfig, _ngxBindInputsService, _ngxBindOutputsService, _ngxBindIODebugService, _parentInjector, _detectorRef) {
        this.viewContainerRef = viewContainerRef;
        this._ngxBindIOConfig = _ngxBindIOConfig;
        this._ngxBindInputsService = _ngxBindInputsService;
        this._ngxBindOutputsService = _ngxBindOutputsService;
        this._ngxBindIODebugService = _ngxBindIODebugService;
        this._parentInjector = _parentInjector;
        this._detectorRef = _detectorRef;
        this.excludeIO = [];
        this.includeIO = [];
        this.excludeInputs = [];
        this.includeInputs = [];
        this.excludeOutputs = [];
        this.includeOutputs = [];
        this.inputs = {
            innerKeys: [],
            hostKeys: []
        };
        this.usedInputs = {};
        this.usedOutputs = {};
        this.destroyed$ = new Subject();
        this.innerSimpleChanges = {};
    }
    /**
     * @param {?} simpleChanges
     * @return {?}
     */
    ngOnChanges(simpleChanges) {
        this.detectComponents();
    }
    /**
     * @return {?}
     */
    ngOnInit() {
        this.detectComponents();
        this.bindAll();
    }
    /**
     * @return {?}
     */
    ngOnDestroy() {
        this.destroyed$.next(true);
        this.destroyed$.complete();
    }
    /**
     * @param {?} innerKey
     * @param {?} value
     * @return {?}
     */
    bindValue(innerKey, value) {
        /** @type {?} */
        const previousValue = this.innerComponent[innerKey];
        this.innerComponent[innerKey] = value;
        if (this._detectorRef && typeof this._detectorRef.markForCheck === 'function') {
            this._detectorRef.markForCheck();
        }
        if (this._detectorRef instanceof ApplicationRef) {
            setTimeout((/**
             * @return {?}
             */
            () => ((/** @type {?} */ (((/** @type {?} */ (this._detectorRef)))))).tick()), 0);
        }
        if (!this._detectorRef) {
            Object.keys(this.innerComponent).forEach((/**
             * @param {?} key
             * @return {?}
             */
            key => {
                if (this.innerComponent[key] instanceof ChangeDetectorRef) {
                    this.innerComponent[key].markForCheck();
                }
            }));
        }
        if (typeof this.innerComponent['ngOnChanges'] === 'function') {
            /** @type {?} */
            const simpleChange = new SimpleChange(previousValue, value, this.innerSimpleChanges[innerKey] === undefined);
            this.innerComponent['ngOnChanges']({ [innerKey]: simpleChange });
            this.innerSimpleChanges[innerKey] = simpleChange;
        }
    }
    /**
     * @return {?}
     */
    bindAll() {
        getBindIOMetadata(this.innerComponent).asInner.manualOutputs = {};
        Object.keys(this.innerComponent)
            .filter((/**
         * @param {?} innerKey
         * @return {?}
         */
        innerKey => this.innerComponent[innerKey] instanceof EventEmitter &&
            ((/** @type {?} */ (this.innerComponent[innerKey]))).observers.length > 0))
            .forEach((/**
         * @param {?} innerKey
         * @return {?}
         */
        innerKey => (getBindIOMetadata(this.innerComponent).asInner.manualOutputs[innerKey] = ((/** @type {?} */ (this.innerComponent[innerKey]))).observers.length)));
        this.inputs = this._ngxBindInputsService.getInputs(this);
        this._ngxBindInputsService.bindInputs(this);
        this._ngxBindInputsService.bindObservableInputs(this);
        this.outputs = this._ngxBindOutputsService.getOutputs(this);
        this._ngxBindOutputsService.bindOutputs(this);
        this._ngxBindIODebugService.showDebugInfo(this, this.debugIsActive());
    }
    /**
     * @return {?}
     */
    detectComponents() {
        var _a, _b;
        if (this.viewContainerRef && !this.innerComponent && !this.hostComponent) {
            // ivy
            this.innerComponent = (_a = ((/** @type {?} */ (this._detectorRef)))) === null || _a === void 0 ? void 0 : _a.context;
            this.hostComponent =
                this._parentInjector &&
                    (((_b = ((/** @type {?} */ (this._parentInjector)))) === null || _b === void 0 ? void 0 : _b._lView) || []).filter((/**
                     * @param {?} item
                     * @return {?}
                     */
                    item => item && item.__ngContext__))[1];
            // if not ivy try old logic
            if (!this.innerComponent) {
                this.innerComponent = this.viewContainerRef['_data'].componentView.component;
            }
            // if not ivy try old logic
            if (!this.hostComponent) {
                this.hostComponent = ((/** @type {?} */ (this.viewContainerRef)))._view.context;
                if (this.hostComponent.$implicit !== undefined) {
                    this.hostComponent = ((/** @type {?} */ (this.viewContainerRef)))._view.component;
                }
            }
        }
    }
    /**
     * @return {?}
     */
    debugIsActive() {
        return this._ngxBindIOConfig.debug || (this.bindIO && this.bindIO.debug);
    }
}
BindIODirective.decorators = [
    { type: Directive, args: [{
                selector: '[bindIO]'
            },] }
];
/** @nocollapse */
BindIODirective.ctorParameters = () => [
    { type: ViewContainerRef },
    { type: undefined, decorators: [{ type: Inject, args: [NGX_BIND_IO_CONFIG,] }] },
    { type: NgxBindInputsService },
    { type: NgxBindOutputsService },
    { type: NgxBindIODebugService },
    { type: Injector, decorators: [{ type: SkipSelf }] },
    { type: ChangeDetectorRef }
];
BindIODirective.propDecorators = {
    bindIO: [{ type: Input }],
    excludeIO: [{ type: Input }],
    includeIO: [{ type: Input }],
    excludeInputs: [{ type: Input }],
    includeInputs: [{ type: Input }],
    excludeOutputs: [{ type: Input }],
    includeOutputs: [{ type: Input }]
};
if (false) {
    /** @type {?} */
    BindIODirective.prototype.bindIO;
    /** @type {?} */
    BindIODirective.prototype.excludeIO;
    /** @type {?} */
    BindIODirective.prototype.includeIO;
    /** @type {?} */
    BindIODirective.prototype.excludeInputs;
    /** @type {?} */
    BindIODirective.prototype.includeInputs;
    /** @type {?} */
    BindIODirective.prototype.excludeOutputs;
    /** @type {?} */
    BindIODirective.prototype.includeOutputs;
    /** @type {?} */
    BindIODirective.prototype.innerComponent;
    /** @type {?} */
    BindIODirective.prototype.hostComponent;
    /** @type {?} */
    BindIODirective.prototype.inputs;
    /** @type {?} */
    BindIODirective.prototype.outputs;
    /** @type {?} */
    BindIODirective.prototype.usedInputs;
    /** @type {?} */
    BindIODirective.prototype.usedOutputs;
    /** @type {?} */
    BindIODirective.prototype.destroyed$;
    /** @type {?} */
    BindIODirective.prototype.innerSimpleChanges;
    /** @type {?} */
    BindIODirective.prototype.ignoreKeysManualBinded;
    /** @type {?} */
    BindIODirective.prototype.viewContainerRef;
    /**
     * @type {?}
     * @private
     */
    BindIODirective.prototype._ngxBindIOConfig;
    /**
     * @type {?}
     * @private
     */
    BindIODirective.prototype._ngxBindInputsService;
    /**
     * @type {?}
     * @private
     */
    BindIODirective.prototype._ngxBindOutputsService;
    /**
     * @type {?}
     * @private
     */
    BindIODirective.prototype._ngxBindIODebugService;
    /**
     * @type {?}
     * @private
     */
    BindIODirective.prototype._parentInjector;
    /**
     * @type {?}
     * @private
     */
    BindIODirective.prototype._detectorRef;
}

/**
 * @fileoverview added by tsickle
 * Generated from: lib/directives/bind-outputs.directive.ts
 * @suppress {checkTypes,constantProperty,extraRequire,missingOverride,missingReturn,unusedPrivateMembers,uselessCode} checked by tsc
 */
class BindOutputsDirective {
    /**
     * @param {?} viewContainerRef
     * @param {?} _ngxBindIOConfig
     * @param {?} _ngxBindOutputsService
     * @param {?} _ngxBindIODebugService
     * @param {?} _parentInjector
     * @param {?} _detectorRef
     */
    constructor(viewContainerRef, _ngxBindIOConfig, _ngxBindOutputsService, _ngxBindIODebugService, _parentInjector, _detectorRef) {
        this.viewContainerRef = viewContainerRef;
        this._ngxBindIOConfig = _ngxBindIOConfig;
        this._ngxBindOutputsService = _ngxBindOutputsService;
        this._ngxBindIODebugService = _ngxBindIODebugService;
        this._parentInjector = _parentInjector;
        this._detectorRef = _detectorRef;
        this.excludeOutputs = [];
        this.includeOutputs = [];
        this.usedOutputs = {};
        this.destroyed$ = new Subject();
    }
    /**
     * @param {?} simpleChanges
     * @return {?}
     */
    ngOnChanges(simpleChanges) {
        this.detectComponents();
    }
    /**
     * @return {?}
     */
    ngOnInit() {
        this.detectComponents();
        this.bindAll();
    }
    /**
     * @return {?}
     */
    bindAll() {
        getBindIOMetadata(this.innerComponent).asInner.manualOutputs = {};
        Object.keys(this.innerComponent)
            .filter((/**
         * @param {?} innerKey
         * @return {?}
         */
        innerKey => this.innerComponent[innerKey] instanceof EventEmitter &&
            ((/** @type {?} */ (this.innerComponent[innerKey]))).observers.length > 0))
            .forEach((/**
         * @param {?} innerKey
         * @return {?}
         */
        innerKey => (getBindIOMetadata(this.innerComponent).asInner.manualOutputs[innerKey] = ((/** @type {?} */ (this.innerComponent[innerKey]))).observers.length)));
        if (this.outputs === undefined) {
            this.outputs = this._ngxBindOutputsService.getOutputs(this);
            this._ngxBindOutputsService.bindOutputs(this);
            this._ngxBindIODebugService.showDebugInfo(this, this.debugIsActive());
        }
    }
    /**
     * @return {?}
     */
    ngOnDestroy() {
        this.destroyed$.next(true);
        this.destroyed$.complete();
    }
    /**
     * @return {?}
     */
    detectComponents() {
        var _a, _b;
        if (this.viewContainerRef && !this.innerComponent && !this.hostComponent) {
            // ivy
            this.innerComponent = (_a = ((/** @type {?} */ (this._detectorRef)))) === null || _a === void 0 ? void 0 : _a.context;
            this.hostComponent =
                this._parentInjector &&
                    (((_b = ((/** @type {?} */ (this._parentInjector)))) === null || _b === void 0 ? void 0 : _b._lView) || []).filter((/**
                     * @param {?} item
                     * @return {?}
                     */
                    item => item && item.__ngContext__))[1];
            // if not ivy try old logic
            if (!this.innerComponent) {
                this.innerComponent = this.viewContainerRef['_data'].componentView.component;
            }
            // if not ivy try old logic
            if (!this.hostComponent) {
                this.hostComponent = ((/** @type {?} */ (this.viewContainerRef)))._view.context;
                if (this.hostComponent.$implicit !== undefined) {
                    this.hostComponent = ((/** @type {?} */ (this.viewContainerRef)))._view.component;
                }
            }
        }
    }
    /**
     * @return {?}
     */
    debugIsActive() {
        return this._ngxBindIOConfig.debug || (this.bindOutputs && this.bindOutputs.debug);
    }
}
BindOutputsDirective.decorators = [
    { type: Directive, args: [{
                selector: '[bindOutputs]'
            },] }
];
/** @nocollapse */
BindOutputsDirective.ctorParameters = () => [
    { type: ViewContainerRef },
    { type: undefined, decorators: [{ type: Inject, args: [NGX_BIND_IO_CONFIG,] }] },
    { type: NgxBindOutputsService },
    { type: NgxBindIODebugService },
    { type: Injector, decorators: [{ type: SkipSelf }] },
    { type: ChangeDetectorRef }
];
BindOutputsDirective.propDecorators = {
    bindOutputs: [{ type: Input }],
    excludeOutputs: [{ type: Input }],
    includeOutputs: [{ type: Input }]
};
if (false) {
    /** @type {?} */
    BindOutputsDirective.prototype.bindOutputs;
    /** @type {?} */
    BindOutputsDirective.prototype.excludeOutputs;
    /** @type {?} */
    BindOutputsDirective.prototype.includeOutputs;
    /** @type {?} */
    BindOutputsDirective.prototype.innerComponent;
    /** @type {?} */
    BindOutputsDirective.prototype.hostComponent;
    /** @type {?} */
    BindOutputsDirective.prototype.outputs;
    /** @type {?} */
    BindOutputsDirective.prototype.usedOutputs;
    /** @type {?} */
    BindOutputsDirective.prototype.destroyed$;
    /** @type {?} */
    BindOutputsDirective.prototype.viewContainerRef;
    /**
     * @type {?}
     * @private
     */
    BindOutputsDirective.prototype._ngxBindIOConfig;
    /**
     * @type {?}
     * @private
     */
    BindOutputsDirective.prototype._ngxBindOutputsService;
    /**
     * @type {?}
     * @private
     */
    BindOutputsDirective.prototype._ngxBindIODebugService;
    /**
     * @type {?}
     * @private
     */
    BindOutputsDirective.prototype._parentInjector;
    /**
     * @type {?}
     * @private
     */
    BindOutputsDirective.prototype._detectorRef;
}

/**
 * @fileoverview added by tsickle
 * Generated from: lib/services/ngx-bind-io.service.ts
 * @suppress {checkTypes,constantProperty,extraRequire,missingOverride,missingReturn,unusedPrivateMembers,uselessCode} checked by tsc
 */
class NgxBindIoService {
    /**
     * @param {?} _ngxBindIOConfig
     * @param {?} _ngxBindInputsService
     * @param {?} _ngxBindOutputsService
     * @param {?} _ngxBindIODebugService
     */
    constructor(_ngxBindIOConfig, _ngxBindInputsService, _ngxBindOutputsService, _ngxBindIODebugService) {
        this._ngxBindIOConfig = _ngxBindIOConfig;
        this._ngxBindInputsService = _ngxBindInputsService;
        this._ngxBindOutputsService = _ngxBindOutputsService;
        this._ngxBindIODebugService = _ngxBindIODebugService;
    }
    /**
     * @param {?} hostComponent
     * @param {?} innerComponent
     * @param {?=} inputs
     * @param {?=} parentInjectorRef
     * @param {?=} changeDetectorRef
     * @param {?=} config
     * @return {?}
     */
    linkHostToInner(hostComponent, innerComponent, inputs, parentInjectorRef, changeDetectorRef, config) {
        /** @type {?} */
        const directive = new BindIODirective(undefined, config
            ? Object.assign(Object.assign({}, this._ngxBindIOConfig), config) : this._ngxBindIOConfig, this._ngxBindInputsService, this._ngxBindOutputsService, this._ngxBindIODebugService, parentInjectorRef, changeDetectorRef);
        directive.ignoreKeysManualBinded = true;
        directive.hostComponent = hostComponent;
        directive.innerComponent = innerComponent;
        if (inputs) {
            /** @type {?} */
            const changes = {};
            Object.keys(inputs).forEach((/**
             * @param {?} innerKey
             * @return {?}
             */
            innerKey => {
                directive.innerComponent[innerKey] = inputs[innerKey];
                changes[innerKey] = new SimpleChange(undefined, inputs[innerKey], true);
            }));
            if (isFunction(directive.innerComponent.ngOnChanges)) {
                directive.innerComponent.ngOnChanges(changes);
            }
        }
        directive.bindAll();
    }
}
NgxBindIoService.decorators = [
    { type: Injectable }
];
/** @nocollapse */
NgxBindIoService.ctorParameters = () => [
    { type: undefined, decorators: [{ type: Inject, args: [NGX_BIND_IO_CONFIG,] }] },
    { type: NgxBindInputsService },
    { type: NgxBindOutputsService },
    { type: NgxBindIODebugService }
];
if (false) {
    /**
     * @type {?}
     * @private
     */
    NgxBindIoService.prototype._ngxBindIOConfig;
    /**
     * @type {?}
     * @private
     */
    NgxBindIoService.prototype._ngxBindInputsService;
    /**
     * @type {?}
     * @private
     */
    NgxBindIoService.prototype._ngxBindOutputsService;
    /**
     * @type {?}
     * @private
     */
    NgxBindIoService.prototype._ngxBindIODebugService;
}

/**
 * @fileoverview added by tsickle
 * Generated from: lib/ngx-bind-io.module.ts
 * @suppress {checkTypes,constantProperty,extraRequire,missingOverride,missingReturn,unusedPrivateMembers,uselessCode} checked by tsc
 */
class NgxBindIOModule {
    /**
     * @param {?=} options
     * @return {?}
     */
    static forRoot(options) {
        return {
            ngModule: NgxBindIOModule,
            providers: [
                {
                    provide: NGX_BIND_IO_CONFIG,
                    useValue: {
                        debug: options ? options.debug : defaultNgxBindIOConfig.debug
                    }
                },
                NgxBindIODebugService,
                NgxBindInputsService,
                NgxBindOutputsService,
                NgxBindIoService
            ]
        };
    }
}
NgxBindIOModule.decorators = [
    { type: NgModule, args: [{
                imports: [CommonModule],
                declarations: [BindInputsDirective, BindOutputsDirective, BindIODirective],
                exports: [BindInputsDirective, BindOutputsDirective, BindIODirective]
            },] }
];

/**
 * @fileoverview added by tsickle
 * Generated from: lib/decorators/bind-io-inner.decorator.ts
 * @suppress {checkTypes,constantProperty,extraRequire,missingOverride,missingReturn,unusedPrivateMembers,uselessCode} checked by tsc
 */
class BindIoInnerLifecycle {
    /**
     * @param {?} simpleChanges
     * @param {?=} runnedFromBindIo
     * @return {?}
     */
    ngOnChanges(simpleChanges, runnedFromBindIo = false) {
        if (!runnedFromBindIo) {
            if (!getBindIOMetadata(this).asInner.manualInputs) {
                getBindIOMetadata(this).asInner.manualInputs = {};
                if (simpleChanges) {
                    Object.keys(simpleChanges).forEach((/**
                     * @param {?} innerKey
                     * @return {?}
                     */
                    innerKey => (getBindIOMetadata(this).asInner.manualInputs[innerKey] = 1)));
                }
            }
            if (((/** @type {?} */ (this))) &&
                ((/** @type {?} */ (this)))[__ORIGINAL_NGONCHANGES__] &&
                typeof ((/** @type {?} */ (this)))[__ORIGINAL_NGONCHANGES__] === 'function') {
                ((/** @type {?} */ (this)))[__ORIGINAL_NGONCHANGES__](simpleChanges, true);
            }
        }
    }
    /**
     * @param {?=} runnedFromBindIo
     * @return {?}
     */
    ngOnDestroy(runnedFromBindIo = false) {
        if (!runnedFromBindIo) {
            Object.keys(getBindIOMetadata(this).asInner.subscriptions).forEach((/**
             * @param {?} subscriptionName
             * @return {?}
             */
            subscriptionName => getBindIOMetadata(this).asInner.subscriptions[subscriptionName].unsubscribe()));
            if (((/** @type {?} */ (this))) &&
                ((/** @type {?} */ (this)))[__ORIGINAL_NGONDESTROY__] &&
                typeof ((/** @type {?} */ (this)))[__ORIGINAL_NGONDESTROY__] === 'function') {
                ((/** @type {?} */ (this)))[__ORIGINAL_NGONDESTROY__](true);
            }
        }
    }
}
/**
 * @return {?}
 */
function BindIoInner() {
    /** @type {?} */
    const func = (/**
     * @param {?} target
     * @return {?}
     */
    (target) => {
        if (!target.prototype[__ORIGINAL_NGONCHANGES__]) {
            /** @type {?} */
            const bindIoInnerLifecycle = new BindIoInnerLifecycle();
            target.prototype[__ORIGINAL_NGONCHANGES__] = target.prototype.ngOnChanges;
            target.prototype[__ORIGINAL_NGONDESTROY__] = target.prototype.ngOnDestroy;
            target.prototype.ngOnChanges = bindIoInnerLifecycle.ngOnChanges;
            target.prototype.ngOnDestroy = bindIoInnerLifecycle.ngOnDestroy;
        }
    });
    return func;
}

/**
 * @fileoverview added by tsickle
 * Generated from: lib/interfaces/ngx-bind-io-config.interface.ts
 * @suppress {checkTypes,constantProperty,extraRequire,missingOverride,missingReturn,unusedPrivateMembers,uselessCode} checked by tsc
 */
/**
 * @record
 */
function INgxBindIOConfig() { }
if (false) {
    /** @type {?} */
    INgxBindIOConfig.prototype.debug;
}

/**
 * @fileoverview added by tsickle
 * Generated from: lib/interfaces/ngx-bind-io-directive.interface.ts
 * @suppress {checkTypes,constantProperty,extraRequire,missingOverride,missingReturn,unusedPrivateMembers,uselessCode} checked by tsc
 */
/**
 * @record
 */
function INgxBindIODirective() { }
if (false) {
    /** @type {?} */
    INgxBindIODirective.prototype.excludeIO;
    /** @type {?} */
    INgxBindIODirective.prototype.includeIO;
    /** @type {?} */
    INgxBindIODirective.prototype.excludeInputs;
    /** @type {?} */
    INgxBindIODirective.prototype.includeInputs;
    /** @type {?} */
    INgxBindIODirective.prototype.excludeOutputs;
    /** @type {?} */
    INgxBindIODirective.prototype.includeOutputs;
    /** @type {?} */
    INgxBindIODirective.prototype.innerComponent;
    /** @type {?} */
    INgxBindIODirective.prototype.hostComponent;
    /** @type {?} */
    INgxBindIODirective.prototype.inputs;
    /** @type {?} */
    INgxBindIODirective.prototype.outputs;
    /** @type {?} */
    INgxBindIODirective.prototype.usedInputs;
    /** @type {?} */
    INgxBindIODirective.prototype.usedOutputs;
    /** @type {?} */
    INgxBindIODirective.prototype.destroyed$;
    /** @type {?|undefined} */
    INgxBindIODirective.prototype.ignoreKeysManualBinded;
    /** @type {?} */
    INgxBindIODirective.prototype.viewContainerRef;
    /**
     * @param {?} key
     * @param {?} value
     * @return {?}
     */
    INgxBindIODirective.prototype.bindValue = function (key, value) { };
}

/**
 * @fileoverview added by tsickle
 * Generated from: lib/interfaces/ngx-bind-io-metadata.interface.ts
 * @suppress {checkTypes,constantProperty,extraRequire,missingOverride,missingReturn,unusedPrivateMembers,uselessCode} checked by tsc
 */
/**
 * @record
 */
function INgxBindIOMetadata() { }
if (false) {
    /** @type {?} */
    INgxBindIOMetadata.prototype.asHost;
    /** @type {?} */
    INgxBindIOMetadata.prototype.asInner;
}

/**
 * @fileoverview added by tsickle
 * Generated from: index.ts
 * @suppress {checkTypes,constantProperty,extraRequire,missingOverride,missingReturn,unusedPrivateMembers,uselessCode} checked by tsc
 */

/**
 * @fileoverview added by tsickle
 * Generated from: public_api.ts
 * @suppress {checkTypes,constantProperty,extraRequire,missingOverride,missingReturn,unusedPrivateMembers,uselessCode} checked by tsc
 */

/**
 * @fileoverview added by tsickle
 * Generated from: ngx-bind-io.ts
 * @suppress {checkTypes,constantProperty,extraRequire,missingOverride,missingReturn,unusedPrivateMembers,uselessCode} checked by tsc
 */

export { BindIODirective, BindInputsDirective, BindIoInner, BindIoInnerLifecycle, BindOutputsDirective, NGX_BIND_IO_CONFIG, NgxBindIODebugService, NgxBindIOModule, NgxBindInputsService, NgxBindIoService, NgxBindOutputsService, __BIND_IO__, __ORIGINAL_NGONCHANGES__, __ORIGINAL_NGONDESTROY__, collectKeys, defaultNgxBindIOConfig, getBindIOMetadata, getPropDescriptor, isFunction, redefineAccessorProperty, redefineSimpleProperty, removeKeysManualBindedInputs, removeKeysManualBindedOutputs, removeKeysNotAllowedConstants, removeKeysUsedInAttributes };
//# sourceMappingURL=ngx-bind-io.js.map
